@@ -1,5 +1,5 @@
 /* 
- * GNetwork: libgnetwork/gnetwork-tcp-server.c
+ * GNetwork Library: libgnetwork/gnetwork-tcp-server.c
  *
  * Copyright (C) 2001, 2003 James M. Cape
  *
@@ -19,14 +19,6 @@
  * USA
  */
 
-/*
-
-GLib 2.0/GObject-based TCP/IP networking sockets wrapper.
-
-Notes on editing:
-	Tab size: 4
-*/
-
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif /* HAVE_CONFIG_H */
@@ -36,6 +28,7 @@ Notes on editing:
 
 #include "gnetwork-tcp-connection.h"
 #include "gnetwork-dns.h"
+#include "gnetwork-interfaces.h"
 #include "gnetwork-threads.h"
 #include "gnetwork-type-builtins.h"
 #include "gnetwork-utils.h"
@@ -60,14 +53,13 @@ Notes on editing:
 enum
 {
   PROP_0,
+
   /* User-settable */
-  ADDRESS,
+  INTERFACE,
+  INTERFACE_INFO,
   PORT,
   REVERSE_LOOKUPS,
 
-  /* GNetworkSslServer Properties */
-  SSL_ENABLED,
-
   /* GNetworkServer Properties */
   SVR_STATUS,
   SVR_BYTES_SENT,
@@ -82,11 +74,12 @@ enum
 struct _GNetworkTcpServerPrivate
 {
   /* Object Properties */
-  gchar *address;
+  gchar *interface;
+  GNetworkInterfaceInfo *interface_info;
   guint port;
-  GSList *connections;
 
   /* GNetworkServer Properties */
+  GSList *connections;
   guint64 bytes_sent;
   guint64 bytes_received;
   guint max_connections;
@@ -95,7 +88,7 @@ struct _GNetworkTcpServerPrivate
   gpointer data;
   GDestroyNotify notify;
 
-  /* The Socket Itself */
+  /* Object Data */
   GIOChannel *channel;
   gint sockfd;
   gint incoming_id;
@@ -109,6 +102,16 @@ struct _GNetworkTcpServerPrivate
 };
 
 
+struct _GNetworkTcpServerCreationData
+{
+  GTypeClass g_class;
+
+  gchar *address;
+  guint16 port;
+  gpointer socket;
+};
+
+
 /* ********************* *
  *  Classwide Variables  *
  * ********************* */
@@ -124,26 +127,26 @@ static GNetworkConnection *
 create_incoming (GNetworkTcpServer * server, const GValue * server_data, gpointer user_data,
 		 GError ** error)
 {
-  const GNetworkTcpServerCreationData * data = g_value_get_boxed (server_data);
+  const GNetworkTcpServerCreationData *data = g_value_get_boxed (server_data);
 
   return g_object_new (GNETWORK_TYPE_TCP_CONNECTION,
 		       "connection-type", GNETWORK_CONNECTION_SERVER,
 		       "address", data->address, "port", data->port,
-		       "socket-fd", data->fd, NULL);
+		       "socket", GPOINTER_TO_INT (data->socket), NULL);
 }
 
 
 static void
-dns_callback (const GNetworkDnsEntry * entry, GNetworkConnection * connection)
+dns_callback (const GSList *entries, const GError * error, GNetworkConnection * connection)
 {
-  if (entry->hostname != NULL && entry->hostname[0] != '\0')
+  if (entries != NULL && entries->data != NULL)
     {
       GNetworkConnectionStatus status;
 
       g_object_get (connection, "status", &status, NULL);
 
       if (status == GNETWORK_CONNECTION_OPEN)
-	g_object_set (connection, "address", entry->hostname, NULL);
+	g_object_set (connection, "address", gnetwork_dns_entry_get_hostname (entries->data), NULL);
     }
 }
 
@@ -181,8 +184,11 @@ cxn_notify_status_cb (GNetworkConnection * connection, GParamSpec * pspec,
 
   if (status == GNETWORK_CONNECTION_CLOSED)
     {
+      g_signal_handlers_disconnect_by_func (connection, cxn_notify_status_cb, server);
+      g_signal_handlers_disconnect_by_func (connection, cxn_received_cb, server);
+      g_signal_handlers_disconnect_by_func (connection, cxn_sent_cb, server);
       server->_priv->connections = g_slist_remove (server->_priv->connections, connection);
-      g_object_unref (G_OBJECT (connection));
+      g_object_unref (connection);
       g_object_notify (G_OBJECT (server), "connections");
     }
 }
@@ -204,26 +210,28 @@ incoming_handler (GIOChannel * channel, GIOCondition cond, GNetworkTcpServer * s
     {
       GNetworkConnection *cxn;
       GError *error;
-      gchar remote_address[INET_ADDRSTRLEN + 1];
       gint cxn_fd;
-      struct sockaddr_in sin;
-      socklen_t sin_size;
+      struct sockaddr *sa;
+      socklen_t sa_size;
       GNetworkTcpServerCreationData creation_data;
       GValue value = { 0 };
 
-      sin_size = sizeof (struct sockaddr_in);
-      cxn_fd = accept (server->_priv->sockfd, (struct sockaddr *) &sin, &sin_size);
+      sa_size = MAX (sizeof (struct sockaddr_in), sizeof (struct sockaddr_in6));
+      sa = g_malloc0 (sa_size);
+      cxn_fd = accept (server->_priv->sockfd, sa, &sa_size);
 
       /* For whatever reason the incoming connection failed */
       if (cxn_fd < 0)
-	return TRUE;
+	{
+	  g_free (sa);
+	  return TRUE;
+	}
 
-      memset (remote_address, 0, INET_ADDRSTRLEN + 1);
-      inet_ntop (AF_INET, &(sin.sin_addr), remote_address, INET_ADDRSTRLEN);
-
-      creation_data.address = remote_address;
-      creation_data.port = g_ntohs (sin.sin_port);
-      creation_data.fd = cxn_fd;
+      creation_data.address = _gnetwork_sockaddr_get_address (sa);
+      creation_data.g_class.g_type = GNETWORK_TYPE_TCP_SERVER_CREATION_DATA;
+      creation_data.port = _gnetwork_sockaddr_get_port (sa);
+      creation_data.socket = GINT_TO_POINTER (cxn_fd);
+      g_free (sa);
 
       g_value_init (&value, GNETWORK_TYPE_TCP_SERVER_CREATION_DATA);
       g_value_set_boxed (&value, &creation_data);
@@ -237,24 +245,18 @@ incoming_handler (GIOChannel * channel, GIOCondition cond, GNetworkTcpServer * s
 	{
 	  /* Add it to the list of connections we've got */
 	  server->_priv->connections = g_slist_prepend (server->_priv->connections,
-							g_object_ref (G_OBJECT (cxn)));
+							g_object_ref (cxn));
 	  if (server->_priv->reverse_lookups)
 	    {
-	      gnetwork_dns_get (remote_address, (GNetworkDnsCallbackFunc) dns_callback,
-				g_object_ref (G_OBJECT (cxn)), g_object_unref, &error);
-
-	      if (error != NULL)
-		{
-		  gnetwork_server_error (GNETWORK_SERVER (server), error);
-		  g_error_free (error);
-		  error = NULL;
-		}
+	      gnetwork_dns_get (creation_data.address, (GNetworkDnsCallbackFunc) dns_callback,
+				g_object_ref (cxn), g_object_unref);
 	    }
 
 	  /* Connect signals & tell the world we've got a new incoming connection */
-	  g_signal_connect (cxn, "received", (GCallback) cxn_received_cb, server);
-	  g_signal_connect (cxn, "sent", (GCallback) cxn_sent_cb, server);
-	  g_signal_connect (cxn, "notify::status", (GCallback) cxn_notify_status_cb, server);
+	  g_signal_connect_object (cxn, "received", G_CALLBACK (cxn_received_cb), server, 0);
+	  g_signal_connect_object (cxn, "sent", G_CALLBACK (cxn_sent_cb), server, 0);
+	  g_signal_connect_object (cxn, "notify::status", G_CALLBACK (cxn_notify_status_cb), server,
+				   0);
 
 	  gnetwork_server_new_connection (GNETWORK_SERVER (server), GNETWORK_CONNECTION (cxn));
 	  g_object_notify (G_OBJECT (server), "connections");
@@ -269,6 +271,7 @@ incoming_handler (GIOChannel * channel, GIOCondition cond, GNetworkTcpServer * s
 	  shutdown (cxn_fd, SHUT_RDWR);
 	  close (cxn_fd);
 	}
+      g_free (creation_data.address);
 
       return TRUE;
     }
@@ -285,8 +288,8 @@ static void
 gnetwork_tcp_server_open (GNetworkTcpServer * server)
 {
   GError *error;
-  struct sockaddr_in sin;
-  gint bind_retval, listen_retval, yes = 1;
+  gint result, value;
+  struct sockaddr_storage sa = { 0 };
 
   g_return_if_fail (GNETWORK_IS_TCP_SERVER (server));
 
@@ -303,18 +306,60 @@ gnetwork_tcp_server_open (GNetworkTcpServer * server)
 
   /* Create the socket */
   errno = 0;
-  server->_priv->sockfd = socket (AF_INET, SOCK_STREAM, 0);
+  server->_priv->sockfd = socket (AF_INET6, SOCK_STREAM, 0);
+  if (server->_priv->sockfd < 0 && errno == EAFNOSUPPORT)
+    {
+      struct sockaddr_in *sin = (struct sockaddr_in *) &sa;
+
+      errno = 0;
+      server->_priv->sockfd = socket (AF_INET, SOCK_STREAM, 0);
+      sa.ss_family = AF_INET;
+      if (server->_priv->interface_info != NULL)
+	{
+	  sin->sin_addr.s_addr =
+	    GNETWORK_IP_ADDRESS32 (gnetwork_interface_info_get_address (server->_priv->interface_info,
+									GNETWORK_PROTOCOL_IPv4), 3);
+	}
+      else
+	{
+	  sin->sin_addr.s_addr = INADDR_ANY;
+	}
+
+      sin->sin_port = g_htons (server->_priv->port);
+    }
+  else
+    {
+      struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) &sa;
+
+      sa.ss_family = AF_INET6;
+      if (server->_priv->interface_info != NULL)
+	{
+	  sin6->sin6_addr =
+	    *((struct in6_addr *) gnetwork_interface_info_get_address (server->_priv->interface_info,
+								       GNETWORK_PROTOCOL_IPv6));
+	}
+      else
+	{
+	  sin6->sin6_addr = in6addr_any;
+	}
+
+      sin6->sin6_port = g_htons (server->_priv->port);
+    }
+
   if (server->_priv->sockfd < 0)
     {
       switch (errno)
 	{
 	case EMFILE:
-	  error = g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES,
-				       gnetwork_server_strerror (GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES));
+	  error =
+	    g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES,
+				 gnetwork_server_strerror
+				 (GNETWORK_SERVER_ERROR_TOO_MANY_PROCESSES));
 	  break;
 	case EACCES:
-	  error = g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_SERVER_TYPE_PERMISSIONS,
-				       gnetwork_server_strerror (GNETWORK_SERVER_ERROR_SERVER_TYPE_PERMISSIONS));
+	  error = g_error_new_literal (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_PERMISSIONS,
+				       gnetwork_server_strerror
+				       (GNETWORK_SERVER_ERROR_PERMISSIONS));
 	  break;
 
 	case ENFILE:
@@ -333,41 +378,35 @@ gnetwork_tcp_server_open (GNetworkTcpServer * server)
       gnetwork_server_error (GNETWORK_SERVER (server), error);
       g_error_free (error);
 
-      server->_priv->status = GNETWORK_SERVER_CLOSED;
-      g_object_notify (G_OBJECT (server), "status");
+      if (server->_priv->status > GNETWORK_SERVER_CLOSED)
+	{
+	  server->_priv->status = GNETWORK_SERVER_CLOSED;
+	  g_object_notify (G_OBJECT (server), "status");
+	}
       return;
     }
 
-  setsockopt (server->_priv->sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int));
   fcntl (server->_priv->sockfd, F_SETFL, O_NONBLOCK);
 
-  /* If we don't have a forced local address, or if the address given isn't an IP
-     address, just use INADDR_ANY */
-  if (server->_priv->address == NULL || inet_aton (server->_priv->address, &(sin.sin_addr)) == 0)
-    {
-      sin.sin_addr.s_addr = INADDR_ANY;
-    }
-  sin.sin_family = AF_INET;
-  sin.sin_port = g_htons (server->_priv->port);
+  value = TRUE;
+  setsockopt (server->_priv->sockfd, SOL_SOCKET, SO_REUSEADDR, &value, sizeof (value));
 
   /* Bind the socket */
   errno = 0;
-  bind_retval = bind (server->_priv->sockfd, ((struct sockaddr *) (&sin)),
-		      sizeof (struct sockaddr_in));
-
-  if (bind_retval < 0)
+  result = bind (server->_priv->sockfd, (struct sockaddr *) &sa, sizeof (sa));
+  if (result < 0)
     {
       switch (errno)
 	{
 	case EINVAL:
-	  error = g_error_new (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_ALREADY_BOUND,
+	  error = g_error_new (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_ALREADY_EXISTS,
 			       _("The service could not be started because there is already "
 				 "another service using port %d."), server->_priv->port);
 	  break;
 	case EACCES:
-	  error = g_error_new (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_SERVER_TYPE_PERMISSIONS,
+	  error = g_error_new (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_PERMISSIONS,
 			       _("You do not have permission to offer a service on port %d. Ports "
-				 "below 1024 can only be used as services by root user."),
+				 "below 1024 can only be used as services by the root user."),
 			       server->_priv->port);
 	  break;
 
@@ -380,23 +419,25 @@ gnetwork_tcp_server_open (GNetworkTcpServer * server)
       gnetwork_server_error (GNETWORK_SERVER (server), error);
       g_error_free (error);
 
-      shutdown (server->_priv->sockfd, SHUT_RDWR);
-      close (server->_priv->sockfd);
-      server->_priv->sockfd = -1;
+      if (server->_priv->status > GNETWORK_SERVER_CLOSED)
+	{
+	  shutdown (server->_priv->sockfd, SHUT_RDWR);
+	  close (server->_priv->sockfd);
+	  server->_priv->sockfd = -1;
 
-      server->_priv->status = GNETWORK_SERVER_CLOSED;
-      g_object_notify (G_OBJECT (server), "status");
+	  server->_priv->status = GNETWORK_SERVER_CLOSED;
+	  g_object_notify (G_OBJECT (server), "status");
+	}
       return;
     }
 
   errno = 0;
-  listen_retval = listen (server->_priv->sockfd, server->_priv->max_connections);
-  if (listen_retval < 0)
+  if (listen (server->_priv->sockfd, server->_priv->max_connections) < 0)
     {
       switch (errno)
 	{
 	case EADDRINUSE:
-	  error = g_error_new (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_ALREADY_BOUND,
+	  error = g_error_new (GNETWORK_SERVER_ERROR, GNETWORK_SERVER_ERROR_ALREADY_EXISTS,
 			       _("The service could not be started because there is already "
 				 "another service using port %d."), server->_priv->port);
 	  break;
@@ -409,22 +450,25 @@ gnetwork_tcp_server_open (GNetworkTcpServer * server)
       gnetwork_server_error (GNETWORK_SERVER (server), error);
       g_error_free (error);
 
-      shutdown (server->_priv->sockfd, SHUT_RDWR);
-      close (server->_priv->sockfd);
-      server->_priv->sockfd = -1;
+      if (server->_priv->status > GNETWORK_SERVER_CLOSED)
+	{
+	  shutdown (server->_priv->sockfd, SHUT_RDWR);
+	  close (server->_priv->sockfd);
+	  server->_priv->sockfd = -1;
 
-      server->_priv->status = GNETWORK_SERVER_CLOSED;
-      g_object_notify (G_OBJECT (server), "status");
+	  server->_priv->status = GNETWORK_SERVER_CLOSED;
+	  g_object_notify (G_OBJECT (server), "status");
+	}
       return;
     }
 
+  server->_priv->status = GNETWORK_SERVER_OPEN;
+  g_object_notify (G_OBJECT (server), "status");
+
   server->_priv->channel = g_io_channel_unix_new (server->_priv->sockfd);
   server->_priv->incoming_id = gnetwork_thread_io_add_watch (server->_priv->channel,
 							     GNETWORK_IO_ANY,
 							     (GIOFunc) incoming_handler, server);
-
-  server->_priv->status = GNETWORK_SERVER_OPEN;
-  g_object_notify (G_OBJECT (server), "status");
 }
 
 
@@ -443,11 +487,27 @@ gnetwork_tcp_server_close (GNetworkTcpServer * server)
        g_slist_remove_link (server->_priv->connections, server->_priv->connections))
     {
       if (server->_priv->close_children)
-	gnetwork_connection_close (GNETWORK_CONNECTION (server->_priv->connections->data));
+	{
+	  gnetwork_connection_close (GNETWORK_CONNECTION (server->_priv->connections->data));
+	}
       else
-	g_object_unref (G_OBJECT (server->_priv->connections->data));
+	{
+	  g_signal_handlers_disconnect_by_func (server->_priv->connections->data,
+						cxn_notify_status_cb, server);
+	  g_signal_handlers_disconnect_by_func (server->_priv->connections->data,
+						cxn_received_cb, server);
+	  g_signal_handlers_disconnect_by_func (server->_priv->connections->data,
+						cxn_sent_cb, server);
+	  g_object_unref (server->_priv->connections->data);
+	}
     }
+  g_object_notify (G_OBJECT (server), "connections");
 
+  if (server->_priv->incoming_id != -1)
+    {
+      gnetwork_thread_source_remove (server->_priv->incoming_id);
+      server->_priv->incoming_id = -1;
+    }
   g_io_channel_shutdown (server->_priv->channel, FALSE, NULL);
   g_io_channel_unref (server->_priv->channel);
   server->_priv->channel = NULL;
@@ -477,7 +537,7 @@ gnetwork_tcp_server_set_create_func (GNetworkTcpServer * server, GNetworkServerC
     {
       server->_priv->create_func = (GNetworkServerCreateFunc) create_incoming;
     }
-  
+
   server->_priv->data = data;
   server->_priv->notify = notify;
 }
@@ -505,32 +565,65 @@ gnetwork_tcp_server_set_property (GObject * object, guint property, const GValue
 
   switch (property)
     {
-    case ADDRESS:
+    case INTERFACE:
       g_return_if_fail (server->_priv->status < GNETWORK_SERVER_OPENING);
 
-      g_free (server->_priv->address);
-      server->_priv->address = g_value_dup_string (value);
+      {
+	GNetworkInterfaceInfo *info;
+	const gchar *interface;
 
-      g_object_notify (object, "address");
+	interface = g_value_get_string (value);
+
+	info = gnetwork_interface_get_info (interface);
+
+	if (info != NULL && info != server->_priv->interface_info)
+	  {
+	    g_free (server->_priv->interface);
+	    server->_priv->interface = g_strdup (server->_priv->interface);
+
+	    gnetwork_interface_info_unref (server->_priv->interface_info);
+	    server->_priv->interface_info = gnetwork_interface_info_ref (info);
+
+	    g_object_notify (object, "interface-info");
+	  }
+
+	gnetwork_interface_info_unref (info);
+      }
+      break;
+    case INTERFACE_INFO:
+      g_return_if_fail (server->_priv->status < GNETWORK_SERVER_OPENING);
+
+      {
+	GNetworkInterfaceInfo *info = g_value_dup_boxed (value);
+
+	if (GNETWORK_IS_INTERFACE_INFO (info))
+	  {
+	    g_free (server->_priv->interface);
+	    server->_priv->interface = g_strdup (gnetwork_interface_info_get_name (info));
+
+	    gnetwork_interface_info_unref (server->_priv->interface_info);
+	    server->_priv->interface_info = gnetwork_interface_info_ref (info);
+
+	    g_object_notify (object, "interface");
+	  }
+
+	gnetwork_interface_info_unref (info);
+      }
       break;
     case PORT:
       g_return_if_fail (server->_priv->status < GNETWORK_SERVER_OPENING);
 
       server->_priv->port = g_value_get_uint (value);
-      g_object_notify (object, "port");
       break;
     case REVERSE_LOOKUPS:
       server->_priv->reverse_lookups = g_value_get_boolean (value);
-      g_object_notify (object, "reverse-lookups");
       break;
 
     case SVR_MAX_CONNECTIONS:
       server->_priv->max_connections = g_value_get_uint (value);
-      g_object_notify (object, "max-connections");
       break;
     case SVR_CLOSE_CHILDREN:
       server->_priv->close_children = g_value_get_boolean (value);
-      g_object_notify (object, "close-children");
       break;
 
     default:
@@ -548,8 +641,11 @@ gnetwork_tcp_server_get_property (GObject * object, guint property, GValue * val
 
   switch (property)
     {
-    case ADDRESS:
-      g_value_set_string (value, server->_priv->address);
+    case INTERFACE:
+      g_value_set_string (value, server->_priv->interface);
+      break;
+    case INTERFACE_INFO:
+      g_value_set_boxed (value, server->_priv->interface_info);
       break;
     case PORT:
       g_value_set_uint (value, server->_priv->port);
@@ -590,6 +686,8 @@ gnetwork_tcp_server_dispose (GObject * object)
   if (server->_priv->status > GNETWORK_SERVER_CLOSED)
     gnetwork_tcp_server_close (server);
 
+  gnetwork_interface_info_unref (server->_priv->interface_info);
+
   if (server->_priv->notify != NULL && server->_priv->data != NULL)
     (*server->_priv->notify) (server->_priv->data);
 
@@ -606,7 +704,8 @@ gnetwork_tcp_server_finalize (GObject * object)
   if (server->_priv->status > GNETWORK_SERVER_CLOSED)
     gnetwork_tcp_server_close (server);
 
-  g_free (server->_priv->address);
+  g_free (server->_priv->interface);
+  g_free (server->_priv);
 
   if (G_OBJECT_CLASS (parent_class)->finalize != NULL)
     (*G_OBJECT_CLASS (parent_class)->finalize) (object);
@@ -629,11 +728,17 @@ gnetwork_tcp_server_class_init (GNetworkTcpServerClass * class)
   object_class->dispose = gnetwork_tcp_server_dispose;
   object_class->finalize = gnetwork_tcp_server_finalize;
 
-  g_object_class_install_property (object_class, ADDRESS,
-				   g_param_spec_string ("address", _("Local Address"),
-							_("The hostname or IP address of the "
-							  "interface to use."), NULL,
+  g_object_class_install_property (object_class, INTERFACE,
+				   g_param_spec_string ("interface", _("Local Interface"),
+							_("The name of the interface to allow "
+							  "connections to (e.g. \"eth0\")."), NULL,
 							(G_PARAM_READWRITE | G_PARAM_CONSTRUCT)));
+  g_object_class_install_property (object_class, INTERFACE_INFO,
+				   g_param_spec_boxed ("interface-info", _("Interface Information"),
+						       _("The hostname or IP address of the "
+							 "interface to allow connections to."),
+						       GNETWORK_TYPE_INTERFACE_INFO,
+						       G_PARAM_READWRITE));
   g_object_class_install_property (object_class, PORT,
 				   g_param_spec_uint ("port", _("Local Port"),
 						      _("The port number to allow connections to."),
@@ -659,10 +764,16 @@ gnetwork_tcp_server_class_init (GNetworkTcpServerClass * class)
 static void
 gnetwork_tcp_server_instance_init (GNetworkTcpServer * server)
 {
-  server->_priv = g_new0 (GNetworkTcpServerPrivate, 1);
+  server->_priv = g_new (GNetworkTcpServerPrivate, 1);
 
+  server->_priv->status = GNETWORK_SERVER_CLOSED;
+  server->_priv->interface = NULL;
+  server->_priv->interface_info = NULL;
   server->_priv->connections = NULL;
 
+  server->_priv->notify = NULL;
+  server->_priv->data = NULL;
+
   server->_priv->bytes_received = 0;
   server->_priv->bytes_sent = 0;
   server->_priv->create_func = (GNetworkServerCreateFunc) create_incoming;
@@ -707,29 +818,51 @@ gnetwork_tcp_server_get_type (void)
 
 /**
  * gnetwork_tcp_server_new:
+ * @interface: the local IP interface to use, or %NULL.
  * @port: the local port to allow connections through.
- * @address: the local IP address to use, or %NULL.
  *
- * Creates a new, unopened server object for incoming connections on
- * @address:@port. See gnetwork_tcp_server_set_address() for more information
- * on how to use the @address parameter.
+ * Creates a new, unopened server object for incoming connections on @port on
+ * @interface. The IP address in @interface should correspond to the interface
+ * that the new server should listen on. If @interface is "127.0.0.1", then only
+ * connections which come in via the loopback (lo) interface are allowed,
+ * whereas if the IP address of eth0 is "192.168.0.10", and @interface is set to
+ * "192.168.0.10", then only connections which come in over eth0 are accepted,
+ * etc. This functions as a bare-bones firewall for this server, though if more
+ * complicated firewalls are required (at this level), a custom
+ * connection-creation function could do the trick.
  *
  * Returns: a new #GNetworkTcpServer object.
  *
  * Since: 1.0
  **/
 GNetworkTcpServer *
-gnetwork_tcp_server_new (guint port, const gchar * address)
+gnetwork_tcp_server_new (const gchar * interface, guint port)
 {
-  struct in_addr addr;
+  GNetworkTcpServer *retval;
 
   g_return_val_if_fail (port <= 65535, NULL);
-  g_return_val_if_fail (address == NULL || inet_aton (address, &addr) != 0, NULL);
+  g_return_val_if_fail (interface == NULL || interface[0] != '\0', NULL);
 
-  return g_object_new (GNETWORK_TYPE_TCP_SERVER, "address", address, "port", port, NULL);
+  if (gnetwork_str_is_ip_address (interface))
+    {
+      GNetworkInterfaceInfo *info = gnetwork_interface_get_info_by_address (interface);
+
+      retval = g_object_new (GNETWORK_TYPE_TCP_SERVER, "interface-info", info, "port", port, NULL);
+      gnetwork_interface_info_unref (info);
+    }
+  else
+    {
+      retval = g_object_new (GNETWORK_TYPE_TCP_SERVER, "interface", interface, "port", port, NULL);
+    }
+
+  return retval;
 }
 
 
+/* ************************************************************************** *
+ *  GNetworkTcpServerCreationData                                             *
+ * ************************************************************************** */
+
 GType
 gnetwork_tcp_server_creation_data_get_type (void)
 {
@@ -758,17 +891,20 @@ gnetwork_tcp_server_creation_data_get_type (void)
  * Since: 1.0
  **/
 GNetworkTcpServerCreationData *
-gnetwork_tcp_server_creation_data_dup (GNetworkTcpServerCreationData * src)
+gnetwork_tcp_server_creation_data_dup (const GNetworkTcpServerCreationData * src)
 {
   GNetworkTcpServerCreationData *dest;
 
+  g_return_val_if_fail (src == NULL || GNETWORK_IS_TCP_SERVER_CREATION_DATA (src), NULL);
+
   if (src == NULL)
     return NULL;
 
   dest = g_new0 (GNetworkTcpServerCreationData, 1);
+  dest->g_class.g_type = GNETWORK_TYPE_TCP_SERVER_CREATION_DATA;
   dest->address = g_strdup (src->address);
   dest->port = src->port;
-  dest->fd = src->fd;
+  dest->socket = src->socket;
 
   return dest;
 }
@@ -785,9 +921,63 @@ gnetwork_tcp_server_creation_data_dup (GNetworkTcpServerCreationData * src)
 void
 gnetwork_tcp_server_creation_data_free (GNetworkTcpServerCreationData * data)
 {
+  g_return_if_fail (data == NULL || GNETWORK_IS_TCP_SERVER_CREATION_DATA (data));
+
   if (data != NULL)
     {
       g_free (data->address);
       g_free (data);
     }
 }
+
+
+/**
+ * gnetwork_tcp_server_creation_data_get_address:
+ * @data: the connection creation data to examine.
+ * 
+ * Retrieves the string of the host IP address of the incoming connection in
+ * @data.
+ * 
+ * Returns: the address of @data.
+ **/
+G_CONST_RETURN gchar *
+gnetwork_tcp_server_creation_data_get_address (const GNetworkTcpServerCreationData * data)
+{
+  g_return_val_if_fail (GNETWORK_IS_TCP_SERVER_CREATION_DATA (data), NULL);
+
+  return data->address;
+}
+
+
+/**
+ * gnetwork_tcp_server_creation_data_get_port:
+ * @data: the connection creation data to examine.
+ * 
+ * Retrieves the @port in @data. This value is 
+ * 
+ * Returns: the address of @data.
+ **/
+guint16
+gnetwork_tcp_server_creation_data_get_port (const GNetworkTcpServerCreationData * data)
+{
+  g_return_val_if_fail (GNETWORK_IS_TCP_SERVER_CREATION_DATA (data), 0);
+
+  return data->port;
+}
+
+
+/**
+ * gnetwork_tcp_server_creation_data_get_socket:
+ * @data: the connection creation data to examine.
+ * 
+ * Retrieves the socket data (a file descriptor on Unix) in @data.
+ * 
+ * Returns: the address of @data.
+ **/
+gconstpointer
+gnetwork_tcp_server_creation_data_get_socket (const GNetworkTcpServerCreationData * data)
+{
+  g_return_val_if_fail (GNETWORK_IS_TCP_SERVER_CREATION_DATA (data), NULL);
+
+  return data->socket;
+}
